{$S-,R-,I-,V-,B-,F-}

{*********************************************************}
{*                   TPEEK70.PAS 7.00                    *}
{*    Copyright (c) 1988,1993 by TurboPower Software.    *}
{*                 All rights reserved.                  *}
{*********************************************************}

unit TPeek70;
  {-Peek at a version 7.0 TPU/TPP/TPW/TPL file}

interface

{$I TADEFINE.INC}

uses
  Dos,
  TData70,
  TMsg70,
  {$IFDEF UseTPro}
  TpString,
  {$ELSE}
  OpString,
  {$ENDIF}
  Unasm;

var
  CurrentUnitSymbol : Word;
  FindTypeNameFunc : function (TP : TpuPointer; P : Pointer; Strict : Boolean) : String;
  FindProcNameByIndexFunc : function (TP : TpuPointer; O : Word) : Boolean;
  ReadSymbolTableProc : procedure (TP : TpuPointer; HashOfst : Word);

function VarClass(vFlags : Byte) : string;
  {-Return the class of a variable, based on its segment type}

function UnitType(Header : TpuPointer) : string;
  {-Returns 'DOS', 'DPMI', or 'Windows'}

function FindTypeClass(TP : TpuPointer; TA : Pointer;
                       var Size : Word; var ST : Byte; var P) : TypeClass;
  {-Return the size and type of a variable, if possible}

function TypeIsObject(PHead : SymHeadPtr) : Boolean;
  {-Return True if P points to an object}

function VarType(TP : TpuPointer; TA : Pointer; var Size : Word) : string;
  {-Return the size and type of a variable, if possible}

procedure DumpRawData;
  {-Dump a 6.0 TPU or TPL file}

function FindSystemSeg(TPU : TpuPointer) : Word;
  {-Find the 'segment' for the SYSTEM unit}

  {==========================================================================}

implementation

  procedure DumpRawType(TP : TpuPointer; P : Pointer; Base : Word); forward;
  procedure DumpSymbol(TP : TpuPointer; PHead : SymHeadPtr; DoCR : Boolean); forward;
  procedure DumpSymbolTable(TP : TpuPointer; ScopePtr : Word); forward;

  function ZeroPad(L : LongInt; W : Byte) : string;
  begin
    ZeroPad := LeftPadCh(Long2Str(L), '0', W);
  end;

  function FileDate2Str(D : LongInt) : string;
    {-Convert a file's date/time to a string of the form 'mm/dd/yy hh:mm'}
  var
    DT : DateTime;
    S : string[40];
  begin
    if D = 0 then
      FileDate2Str := '              '
    else begin
      UnpackTime(D, DT);
      with DT do begin
        S := ZeroPad(Month, 2)+'/'+ZeroPad(Day, 2)+'/'+ZeroPad(Year-1900, 2);
        FileDate2Str := S+' '+ZeroPad(Hour, 2)+':'+ZeroPad(Min, 2);
      end;
    end;
  end;

  function SymbolAt(TP : TpuPointer; O : Word) : string;
    {-Return the name of the symbol that starts at offset O}
  var
    SHP : SymHeadPtr;
  begin
    SHP := Ptr(SegOfs(TP).Segm, O);
    SymbolAt := SHP^.SymName;
  end;

  procedure DumpWord(Ofst, W : Word; S : string);
    {-Dump a word value in a TPU header}
  begin
    WriteLn(HexW(Ofst), ' ', HexW(W), '   ', S);
  end;

  procedure DumpLabel(P : Pointer);
    {-Dump a label symbol}
  var
    LR : ^LabelRec absolute P;
  begin
    with LR^ do begin
      WriteLn(HexW(Ofs(aTarget)),   ' ', HexW(aTarget), ' target');
      WriteLn(HexW(Ofs(aJumpList)), ' ', HexW(aJumpList), ' jump list');
    end;
  end;

  procedure DumpSystemProc(P : Pointer);
    {-Dump a system procedure/function symbol}
  var
    SP : ^SysProcRec absolute P;
  begin
    with SP^ do
      WriteLn(HexW(Ofs(JumpTableIndex)), ' ', HexW(JumpTableIndex),
        ' compiler jump table index');
  end;

  procedure DumpPortInst(P : Pointer);
    {-Dump a PORT or PORTW symbol}
  var
    PR : ^PortRec absolute P;
  begin
    with PR^ do
      WriteLn(HexW(Ofs(poSize)), ' ', HexB(poSize), '   size');
  end;

  function VarClass(vFlags : Byte) : string;
    {-Return the class of a variable, based on its segment type}
  begin
    if vFlags and vAlias <> 0 then
      VarClass := 'abs'
    else case vFlags and $0F of
      vData       : VarClass := 'var';
      vConst      : VarClass := 'const';
      vStack      : VarClass := 'local';
      vAbsolute   : VarClass := 'abs';
      vStack+vVar : VarClass := 'var param';
      vField      : VarClass := 'field';
      else          VarClass := '?';
    end;
  end;

  function FindTypeClass(TP : TpuPointer; TA : Pointer; {**}
                         var Size : Word; var ST : Byte; var P) : TypeClass;
    {-Return the size and type of a variable, if possible}
  var
    MT : TypeClass;
    TASO : SegOfs absolute TA;
    PT : ^RawTypeRec absolute P;
    PTSO : SegOfs absolute PT;
  begin
    ST := 0;
    Size := 0;
    PT := nil;
    if TA = nil then begin
      FindTypeClass := tUntyped;
      Exit;
    end
    else
      MT := tUnknown;

    {find MT and ST, if possible}
    if TASO.Segm = CurrentUnitSymbol then begin
      PT := TA;
      PTSO.Segm := Seg(TP^);
      with PT^ do begin
        MT := tClass;
        ST := tFlags;
        Size := tSize;
        if TASO.Segm = SystemSeg then
          if (MT = tPointer) and (TASO.Ofst = $01EE) then
            ST := $FF;
      end;
    end
    else if TASO.Segm = SystemSeg then begin
      {it's in SYSTEM}
      case TASO.Ofst of
        $0082 :
          begin MT := tUntyped; Size := 0; end;
        $008A :                   {file}
          begin MT := tFile; Size := $80; end;
        $0096 :                   {text}
          begin MT := tText; Size := $100; end;
        $00AE :                   {pointer}
          begin MT := tPointer; Size := 4; end;
        $00BA :                   {string}
          begin MT := tString; Size := $100; end;
        $00DA :                   {single}
          begin MT := tFloat; ST := rSingle; Size := 4; end;
        $00E2 :                   {double}
          begin MT := tFloat; ST := rDouble; Size := 8; end;
        $00EA :                   {comp}
          begin MT := tFloat; ST := rComp; Size := 8; end;
        $00F2 :                   {extended}
          begin MT := tFloat; ST := rExtended; Size := 10; end;
        $0102 :                   {shortint}
          begin MT := tInteger; ST := iShortInt; Size := 1; end;
        $012A :                   {integer}
          begin MT := tInteger; ST := iInteger; Size := 2; end;
        $0152 :                   {longint}
          begin MT := tInteger; ST := iLongInt; Size := 4; end;
        $0116 :                   {byte}
          begin MT := tInteger; ST := iByte; Size := 1; end;
        $013E :                   {word}
          begin MT := tInteger; ST := iWord; Size := 2; end;
        $0172 :                   {boolean}
          begin MT := tBoolean; ST := bBoolean; Size := 1; end;
        $0186 :                   {bytebool}
          begin MT := tBoolean; ST := bByteBool; Size := 1; end;
        $019A :                   {wordbool}
          begin MT := tBoolean; ST := bWordBool; Size := 2; end;
        $01AE :                   {longbool}
          begin MT := tBoolean; ST := bLongBool; Size := 4; end;
        $01CE :                   {char}
          begin MT := tChar; Size := 1; end;
        $00FA :                   {real}
          begin MT := tReal; Size := 6; end;
        $01EE :                   {pchar}
          begin MT := tPointer; ST := $FF; Size := 4; end;
        else
          writeln('?? (TA = ',hexptr(TA), ')');
      end;
    end;

    FindTypeClass := MT;
  end;

  function VarType(TP : TpuPointer; TA : Pointer; var Size : Word) : string;
    {-Return the size and type of a variable, if possible}
  var
    MT : TypeClass;
    ST : Byte;
    PT : ^RawTypeRec;
  begin
    MT := FindTypeClass(TP, TA, Size, ST, PT);

    case MT of
      tUntyped :
        VarType := '?';
      tBoolean :
        case ST of
          bBoolean  : VarType := 'Boolean';
          bByteBool : VarType := 'ByteBool';
          bWordBool : VarType := 'WordBool';
          bLongBool : VarType := 'LongBool';
        end;
      tFloat :
        case ST of
          rSingle   : VarType := 'Single';
          rExtended : VarType := 'Extended';
          rDouble   : VarType := 'Double';
          rComp     : VarType := 'Comp';
          else        VarType := 'unknown';
        end;
      tInteger :
        case ST of
          iShortInt : VarType := 'ShortInt';
          iByte     : VarType := 'Byte';
          iInteger  : VarType := 'Integer';
          iWord     : VarType := 'Word';
          iLongInt  : VarType := 'LongInt';
          else        VarType := 'subrange';
        end;
      tPointer :
        case ST of
          $FF : VarType := 'PChar';
          else VarType := ClassTypes[MT];
        end;
      else
        VarType := ClassTypes[MT];
    end;
  end;

  procedure DumpConstant(TP : TpuPointer; P : Pointer);
    {-Dump a constant symbol}
  var
    CR : ^ConstRec absolute P;
    PSO : SegOfs absolute P;
  begin
    with CR^ do begin
      WriteLn(HexW(PSO.Ofst+1), ' ', HexPtr(cType), ' type def (',
        FindTypeNameFunc(TP, cType, False), ')');

      {following doesn't handle real constants!}
      WriteLn(HexW(PSO.Ofst+3), ' ', HexL(cValue), '  value');

      DumpRawType(TP, cType, PSO.Segm);
    end;
  end;

  function ParamType(Flags : Byte) : string;
  begin
    case Flags of
      vStack :
        ParamType := 'value param';
      vStack+vVar :
        ParamType := 'var param'
      else
        ParamType := 'unknown'
    end;
  end;

  function GetObjectLink(TP : TpuPointer; O : Word) : string;
  var
    ORP : ^ObjectRec;
  begin
    with TP^ do
      if O = 0 then
        GetObjectLink := '*none*'
      else begin
        ORP := Ptr(Seg(TP^), O);
        if ORP^.tNameOfs = 0 then
          GetObjectLink := '??'
        else
          GetObjectLink := SymbolAt(TP, ORP^.tNameOfs);
      end;
  end;

  procedure DumpRawType {(TP : TpuPointer; P : Pointer; Base : Word)} ;
    {-Dump a low-level type definition}
  var
    SaveP : Pointer;
    TR : RawTypePtr absolute P;
    TA : ^ArrayRec absolute P;
    TX : ^TextRec absolute P;
    TRR : ^RecordRec absolute P;
    TPF : ^ProcFuncRec absolute P;
    TPO : ^ObjectRec absolute P;
    PSO : SegOfs absolute P;
    PHead : SymHeadPtr;
    Class : TypeClass;
    PName : ^String;
    I : Word;

    function InterpretSubType(Main : TypeClass; Sub : Byte) : string;
      {-Return a string corresponding to a subtype}
    begin
      if Main = tFloat then
        case Sub of
          rSingle   : InterpretSubType := 'single';
          rExtended : InterpretSubType := 'extended';
          rDouble   : InterpretSubType := 'double';
          rComp     : InterpretSubType := 'comp';
          else        InterpretSubType := 'unknown';
        end
      else
        if Main = tInteger then
          case Sub of
            iShortInt : InterpretSubType := 'shortint';
            iByte     : InterpretSubType := 'byte';
            iInteger  : InterpretSubType := 'integer';
            iWord     : InterpretSubType := 'word';
            iLongInt  : InterpretSubType := 'longint';
            else        InterpretSubType := 'sub-range';
          end
        else if (Main = tBoolean) then
          case Sub of
            bBoolean  : InterpretSubType := 'boolean';
            bByteBool : InterpretSubType := 'bytebool';
            bWordBool : InterpretSubType := 'wordbool';
            bLongBool : InterpretSubType := 'longbool';
          end
        else if (Main = tChar) then
          {char}
          case Sub of
            dSigned   : InterpretSubType := 'signed';
            dUnsigned : InterpretSubType := 'unsigned';
            else        InterpretSubType := '(unknown)';
          end
        else if (Main = tPointer) then
          {pointer}
          case Sub of
            iLongInt : InterpretSubType := 'allow cast to longint';
            else       InterpretSubType := '(unknown)';
          end
        else
          InterpretSubType := '?';
    end;

    procedure DumpRecordFields(First, TableSeg, TableOfst : Word);
      {-Dump a linked series of record fields}
    var
      PH : SymHeadPtr;
      PFR : ^FieldRec;
      O : Word;
      Table : ^LinkTable;
    begin
      Table := Ptr(TableSeg, TableOfst);
      with Table^ do begin
        WriteLn(HexW(TableOfst), ' ', HexW(HashMask), ' hash mask');
        WriteLn(
          HexW(TableOfst+2), ' ',
          HexW(Links[1]), ' ',
          HexW(Links[2]), ' ',
          HexW(Links[3]), ' ',
          HexW(Links[4]), '  links to record fields');
        O := First;
        while O <> 0 do begin
          WriteLn('---');
          PH := Ptr(TableSeg, O);
          PFR := Ptr(TableSeg, O+Length(PH^.SymName)+4);
          DumpSymbol(Ptr(TableSeg, 0), PH, False);
          O := PFR^.vNext;
        end;
      end;
    end;

    procedure DumpObjectFields(First, TableSeg, TableOfst : Word);
      {-Dump a linked series of object fields/methods}
    var
      PH : SymHeadPtr;
      PFR : ^FieldRec;
      PRP : ^ProcRec absolute PFR;
      O : Word;
      Table : ^LinkTable;
    begin
      Table := Ptr(TableSeg, TableOfst);
      with Table^ do begin
        WriteLn(HexW(TableOfst), ' ', HexW(HashMask), ' hash mask');
        WriteLn(
          HexW(TableOfst+2), ' ',
          HexW(Links[1]), ' ',
          HexW(Links[2]), ' ',
          HexW(Links[3]), ' ',
          HexW(Links[4]), ' ',
          HexW(Links[5]), ' ',
          HexW(Links[6]), ' ',
          HexW(Links[7]), ' ',
          HexW(Links[8]), '  links to fields/methods');
        O := First;
        while O <> 0 do begin
          WriteLn('---');
          PH := Ptr(TableSeg, O);
          PFR := Ptr(TableSeg, O+Length(PH^.SymName)+4);
          DumpSymbol(Ptr(TableSeg, 0), PH, False);
          if Char(Byte(PH^.SymCode) AND $7F) = 'Q' then
            O := PFR^.vNext
          else
            O := PRP^.pType.tNext;
        end;
      end;
    end;

  begin
    {get out if the type is in another unit}
    if (PSO.Segm <> CurrentUnitSymbol) then
      Exit;

    {get out if P is nil or we're recurring dangerously}
    if (P = nil) or Recurring(P) then
      Exit;

    {push this address onto the recursion stack}
    PushAddr(P);

    SaveP := P;
    WriteLn(HexPtr(P), ':');
    PSO.Segm := BaseSeg+MemW[Base:PSO.Segm];
    with TR^ do begin
      WriteLn(HexW(Ofs(tClass)), ' ', HexB(Ord(tClass)), '   ', ClassTypes[tClass]);
      WriteLn(HexW(Ofs(tFlags)), ' ', HexB(tFlags),      '   ', InterpretSubType(tClass, tFlags));
      WriteLn(HexW(Ofs(tSize)),  ' ', HexW(tSize),       ' size');
      Class := tClass;
    end;

    case Class of
      tArray,                {array}
      tString :              {string}
        with TA^ do begin
          WriteLn(HexW(Ofs(tComp)), ' ', HexPtr(tComp), ' component (',
            FindTypeNameFunc(TP, tComp, True), ')');
          WriteLn(HexW(Ofs(tIndex)), ' ', HexPtr(tIndex), ' index (',
            FindTypeNameFunc(TP, tIndex, True), ')');
          DumpRawType(TP, tComp, PSO.Segm);
          DumpRawType(TP, tIndex, PSO.Segm);
        end;
      tPointer,              {pointer (subtype is 0D)}
      tSet,                  {set}
      tFile,                 {file (typed or untyped)}
      tText :                {text}
        with TX^ do
          WriteLn(HexW(Ofs(tComp)), ' ', HexPtr(tComp), ' component (',
            FindTypeNameFunc(TP, tComp, True), ')');
      tScalar :              {enumerated}
        with TR^ do begin
          WriteLn(HexW(Ofs(tMinVal)), ' ', HexL(tMinVal),  '  low');
          WriteLn(HexW(Ofs(tMaxVal)), ' ', HexL(tMaxVal),  '  high');
          WriteLn(HexW(Ofs(tBase)),   ' ', HexPtr(tBase),  ' base type def');
          WriteLn(HexW(Ofs(tClass2)), ' ', HexB(tClass2),  ' enumerated set');
          WriteLn(HexW(Ofs(tFlags2)), ' ', HexB(tFlags2),  ' set flags');
          WriteLn(HexW(Ofs(tSize2)),  ' ', HexL(tSize2),   ' set size');
          WriteLn(HexW(Ofs(tBase2)),  ' ', HexPtr(tBase2), ' set base type def');
          WriteLN(HexW(Ofs(tFiller)), ' ', HexW(tFiller),  ' ??');

          {if this is the typedef, a series of constants follow}
          if tBase = SaveP then begin
            PHead := @tNextField;
              for I := 0 to (tMaxVal-tMinVal) do begin
              DumpSymbol(Ptr(Seg(TR^), 0), PHead, False);
              Inc(SegOfs(PHead).Ofst, 12+Length(PHead^.SymName));
            end;
          end;
        end;
      tRecord :              {record}
        with TRR^ do begin
          WriteLn(HexW(Ofs(tParent)), ' ', HexW(tParent),  ' offset of parent');
          WriteLn(HexW(Ofs(tDunno)), ' ', HexW(tDunno),  ' ??');
          WriteLn(HexW(Ofs(tScope)), ' ', HexW(tScope),  ' scope pointer');
          WriteLn(HexW(Ofs(tFieldPtr)), ' ', HexW(tFieldPtr), ' first field');
          DumpRecordFields(tFieldPtr, PSO.Segm, tScope);
        end;
      tProc :                {procedure/function}
        with TPF^ do begin
          Write(HexW(Ofs(tResType)), ' ', HexPtr(tResType), ' result type');
          if tResType = nil then
            WriteLn
          else
            WriteLn(' (', FindTypeNameFunc(TP, tResType, True), ')');
          WriteLn(HexW(Ofs(tParCnt)),  ' ', HexW(tParCnt),    ' param count');
          DumpRawType(TP, tResType, PSO.Segm);
          if tParCnt <> 0 then begin
            WriteLn('Ofst TypeDef   Flags ParamType');
            for I := 1 to tParCnt do
              with tParams[I] do
                WriteLn(HexW(Ofs(rType)), ' ', HexPtr(rType), ' ', HexB(rFlags),
                  '    ', FindTypeNameFunc(TP, rType, False), ' (', ParamType(rFlags), ')');
          end;
        end;
      tUntyped,              {untyped}
      tFloat,                {8087 real}
      tReal : ;              {6-byte real}

      tObject :              {object}
        with TPO^ do begin
          PName := Ptr(PSO.Segm, tNameOfs+3);
          WriteLn(HexW(Ofs(tNameOfs)),   ' ', HexW(tNameOfs),   ' ofs. of object name (',
            PName^, ')');
          WriteLn(HexW(Ofs(tHuh)),     ' ', HexW(tHuh),     ' ??');
          WriteLn(HexW(Ofs(tScope)),     ' ', HexW(tScope),     ' scope pointer');
          Write(HexW(Ofs(tFieldPtr)),  ' ', HexW(tFieldPtr),  ' first symbol');
          if tFieldPtr = 0 then
            Writeln
          else
            WriteLn(' (-> ', SymbolAt(TP, tFieldPtr), ')');
          Write(HexW(Ofs(tParent)),    ' ', HexPtr(tParent),  ' parent type');
          if tParent = nil then
            Writeln
          else
            WriteLn(' (', FindTypeNameFunc(TP, tParent, True), ')');
          WriteLn(HexW(Ofs(tMethSize)),  ' ', HexW(tMethSize),  ' VMT size');
          WriteLn(HexW(Ofs(tMethGroup)), ' ', HexW(tMethGroup), ' const group index for VMT');
          WriteLn(HexW(Ofs(tMethOfs)),   ' ', HexW(tMethOfs),   ' ofs. in object of VMT ptr');
          WriteLn(HexW(Ofs(tDMTGroup)),  ' ', HexW(tDMTGroup),  ' const group index for DMT');
          WriteLn(HexW(Ofs(tPrevObj)),   ' ', HexW(tPrevObj),   ' previous object in linked list (',
            GetObjectLink(TP, tPrevObj), ')');
          WriteLn(HexW(Ofs(tHuh1)),      ' ', HexW(tHuh1),      ' ??');
          WriteLn(HexW(Ofs(tHuh2)),      ' ', HexW(tHuh2),      ' ??');
          WriteLn(HexW(Ofs(tHuhPtr1)),   ' ', HexPtr(tHuhPtr1), ' ??');
          WriteLn(HexW(Ofs(tHuhPtr2)),   ' ', HexPtr(tHuhPtr2), ' ??');

          DumpObjectFields(tFieldPtr, PSO.Segm, tScope);
        end;
      else
        with TR^ do begin
          WriteLn(HexW(Ofs(tMinVal)), ' ', HexL(tMinVal), '  low');
          WriteLn(HexW(Ofs(tMaxVal)), ' ', HexL(tMaxVal), '  high');
          WriteLn(HexW(Ofs(tBase)),   ' ', HexPtr(tBase), ' base type');
        end;
    end;

    {take this address off the recursion stack}
    PopAddr;
  end;

  procedure DumpType(TP : TpuPointer; P : Pointer);
    {-Dump a type symbol}
  var
    TR : ^TypeRec absolute P;
    PSO : SegOfs absolute P;
  begin
    with TR^ do begin
      WriteLn(HexW(PSO.Ofst), ' ', HexPtr(tsType), ' type def');
      DumpRawType(TP, tsType, PSO.Segm);
    end;
  end;

  procedure DumpVariable(TP : TpuPointer; P : Pointer);
    {-Dump a variable symbol}
  var
    VR : VarRecPtr absolute P;
    FR : ^FieldRec absolute P;
    PSO : SegOfs absolute P;

    function InterpretSegment(vFlags : Byte) : string;
      {-Interpret one-byte segment code}
    begin
      if vFlags and vAlias <> 0 then
        InterpretSegment := 'absolute var'
      else case vFlags and $0F of
        vData       : InterpretSegment := 'global var';
        vConst      : InterpretSegment := 'typed constant';
        vStack      : InterpretSegment := 'local';
        vAbsolute   : InterpretSegment := 'absolute var';
        vStack+vVar : InterpretSegment := 'var param';
        vField      : InterpretSegment := 'field';
        else          InterpretSegment := '?';
      end;
    end;

  begin
    with VR^ do begin
      WriteLn(HexW(Ofs(vFlags)), ' ', HexB(vFlags), '   ',
        InterpretSegment(vFlags));

      {handle record fields differently than variables}
      if vFlags = vField then
        with FR^ do begin
          WriteLn(HexW(Ofs(vOffset)), ' ', HexW(vOffset), ' rel. offset within parent');
          WriteLn(HexW(Ofs(vParent)), ' ', HexW(vParent), ' offset of parent data type');
          Write(HexW(Ofs(vNext)), ' ', HexW(vNext), ' next field');
          if vNext = 0 then
            Writeln
          else
            WriteLn(' (-> ', SymbolAt(TP, vNext), ')');
        end
      else begin
        WriteLn(HexW(Ofs(vAddress)), ' ', HexPtr(vAddress), ' address');
        WriteLn(HexW(Ofs(vHuh)), ' ', HexW(vHuh), ' ??');
      end;
      WriteLn(HexW(Ofs(vType)), ' ', HexPtr(vType), ' type def (',
        FindTypeNameFunc(TP, vType, True), ')');

      DumpRawType(TP, vType, PSO.Segm);
    end;
  end;

  function FindCodeAddress(TP : TpuPointer; ProcTableIndex : Word) : Pointer;
    {-Given an index into the procedure table, return the address shown in
      a MAP file}
  var
    CBP : ProcTablePtr;
    LRP : ^GroupRec;
    I, Stop : Word;
  begin
    SegOfs(CBP).Segm := Seg(TP^);
    SegOfs(CBP).Ofst := TP^.ProcPtr+ProcTableIndex;
    if CBP^.prFlags and ptExtern <> 0 then begin
      FindCodeAddress := Ptr($FFFF, $FFFF);
      Exit;
    end;
    SegOfs(LRP).Segm := Seg(TP^);
    SegOfs(LRP).Ofst := TP^.GroupPtr;
    Stop := SegOfs(LRP).Ofst+CBP^.prGroup;

    I := 0;
    while SegOfs(LRP).Ofst < Stop do begin
      Inc(I, LRP^.grCodeSize);
      Inc(SegOfs(LRP).Ofst, SizeOf(GroupRec));
    end;

    {actual offset is the starting offset for the entire block, plus the
     amount of constant data at the head of the block. If the "block" is
     an entire OBJ file, "prOffset" includes both code and data.}
    FindCodeAddress := Ptr(0, I+CBP^.prOffset);
  end;

  function DumpHashTable(TP : TpuPointer; ScopePtr : Word) : Boolean;
    {-Dump the hash table}
  var
    P : ^string;
    PSO : SegOfs absolute P;
    UnitSeg, Entries, Hash, Index, I, O : Word;
    Empty : Boolean;
  begin
    DumpHashTable := True;

    UnitSeg := Seg(TP^);
    P := Pointer(TP);
    O := ScopePtr;
    Hash := MemW[UnitSeg:O];
    Entries := Hash shr 1;
    if SymNest <> 0 then
      WriteLn('Local Hash Table, level ', SymNest, ' (unused entries not shown)')
    else
      WriteLn('Hash Table (unused entries not shown)'^M^J,
       '-------------------------------------');
    WriteLn(HexW(O), ' ', HexW(Hash), ' Hash mask');

    Empty := True;
    for I := 0 to Entries do begin
      Inc(O, 2);
      if MemW[UnitSeg:O] <> 0 then
        Empty := False;
    end;
    if Empty then begin
      WriteLn('(empty)');
      DumpHashTable := False;
      Exit;
    end;

    WriteLn(^M^J'Ofst Val  Points to'^M^J'---- ---- ---------');
    O := ScopePtr;
    for I := 0 to Entries do begin
      Inc(O, 2);
      Index := MemW[UnitSeg:O];
      PSO.Ofst := Index+3;
      if Index <> 0 then
        WriteLn(HexW(O), ' ', HexW(Index), ' ', P^);
    end;
    WriteLn;
  end;

  procedure DumpProcedure(TP : TpuPointer; P : Pointer);
    {-Dump a procedure symbol}
  var
    PR : ^ProcRec absolute P;
    ICP : ^Byte;
    ICPSO : SegOfs absolute ICP;
    PA : ^Array4 absolute ICP;
    CA : Pointer;
    I, Stop : Word;
    B : Byte;
    IsMacro : Boolean;
    S : string[32];
  begin
    with PR^, pType do begin
      IsMacro := (pFlags = pInline);
      case pFlags and not (pExternal+pAssembler) of
        pNear        : S := 'near';
        pFar         : S := 'far';
        pInline      : S := 'inline macro';
        pInterrupt   : S := 'interrupt';
        pFar+pMethod : S := 'method';
        pFar+pMethod+pConstruct : S := 'constructor';
        pFar+pMethod+pDestruct  : S := 'destructor';
        else S := '';
      end;
      if pFlags and pExternal <> 0 then
        S := S + '; external';
      if pFlags and pAssembler <> 0 then
        S := S + '; assembler';
      WriteLn(HexW(Ofs(pFlags)), ' ', HexB(pFlags), '   ', S);
      Write(HexW(Ofs(pCodeGenFlags)), ' ', HexB(pCodeGenFlags), '   ');
      case pCodeGenFlags and cgStackMask of
        cgNoFrame :
          Write('no stack frame');
        cgHasFrame :
          Write('normal stack frame');
        cgWinFrame+cgHasFrame :
          Write('Windows stack frame');
      end;
      if pCodeGenFlags and cgExport <> 0 then
        Write(', exported');
      if pCodeGenFlags and cgDMT <> 0 then
        Write(', use DMT');
      WriteLn;
      Write(HexW(Ofs(pIndex)), ' ', HexW(pIndex));
      if IsMacro then
        WriteLn(' size of code/relocation data block')
      else begin
        CA := FindCodeAddress(TP, pIndex);
        Write(' index into procedure table');
        if CA <> Ptr($FFFF, $FFFF) then
          Write(' (', HexPtr(CA), ')');
        WriteLn;
      end;
      WriteLn(HexW(Ofs(pParent)), ' ', HexW(pParent), ' offset of parent data type');
      WriteLn(HexW(Ofs(pScope)), ' ', HexW(pScope),
        ' offset of local hash table');
      Write(HexW(Ofs(pMethOfs)), ' ', HexW(pMethOfs));
      if pFlags and pMethod = 0 then
        WriteLn(' unused')
      else begin
        if pCodeGenFlags and cgDMT <> 0 then begin
          Write(' DMT index [');
          if TP^.UnitFlags and ufWindows = ufWindows then begin
            if pMethOfs >= $A000 then
              Write('cm_First + ', CMtoStr(pMethOfs-$A000))
            else if pMethOfs >= $9000 then
              Write('nf_First + ', pMethOfs-$9000)
            else if pMethOfs >= $8000 then
              Write('id_First + ', IDtoStr(pMethOfs-$8000))
            else
              Write('wm_First + ', WMtoStr(pMethOfs));
          end
          else
            Write(pMethOfs);
          WriteLn(']');
        end
        else if pMethOfs = 0 then
          WriteLn(' static method')
        else
          WriteLn(' VMT offset');
      end;
      WriteLn(HexW(Ofs(tClass)), ' ', HexB(Ord(tClass)), '   ', ClassTypes[tClass]);
      WriteLn(HexW(Ofs(tFlags)), ' ', HexB(tFlags), '   type flags');
      WriteLn(HexW(Ofs(tSize)),  ' ', HexW(tSize),  ' type size');
      WriteLn(HexW(Ofs(tHuh)),  ' ', HexW(tHuh), ' ??');
      Write(HexW(Ofs(tNext)),  ' ', HexW(tNext), ' offset of next method');
      if tNext = 0 then
        Writeln
      else
        WriteLn(' (-> ', SymbolAt(TP, tNext), ')');
      Write(HexW(Ofs(tResType)), ' ', HexPtr(tResType), ' result type');
      if tResType = nil then
        WriteLn
      else
        WriteLn(' (', FindTypeNameFunc(TP, tResType, True), ')');
      WriteLn(HexW(Ofs(tParCnt)), ' ', HexW(tParCnt), ' # of params');

      if tParCnt <> 0 then begin
        WriteLn('Ofst TypeDef   Flags ParamType');
        for I := 1 to tParCnt do
          with tParams[I] do
            WriteLn(HexW(Ofs(rType)), ' ', HexPtr(rType), ' ', HexB(rFlags),
              '    ', FindTypeNameFunc(TP, rType, False), ' (', ParamType(rFlags), ')');
      end;

      if IsMacro then begin
        ICP := @tParams[tParCnt+1];
        Stop := ICPSO.Ofst+pIndex;
        WriteLn('Ofst RD Code (data if RD<>0)');
        while ICPSO.Ofst < Stop do begin
          B := ICP^;
          Write(HexW(ICPSO.Ofst), ' ', HexB(B));
          case B of
            0 :
              begin
                WriteLn(' ', HexB(PA^[1]));
                Inc(ICPSO.Ofst, 2);
              end;
            else
              begin
                for I := 1 to B+2 do
                  Write(' ', HexB(PA^[I]));
                WriteLn;
                Inc(ICPSO.Ofst, 3+B);
              end;
          end;
        end;
      end else if (pScope <> 0) then begin
        {Show local symbol table}
        Inc(SymNest);
        if DumpHashTable(TP, pScope) then begin
          DumpSymbolTable(TP, pScope);
          WriteLn(
            '----------------------------------- End local symbols level ',
            SymNest);
        end;
        Dec(SymNest);
      end;
    end;
  end;

  procedure DumpMemInst(TP : TpuPointer; P : Pointer);
    {-Dump a MEM/MEMW/MEML symbol}
  var
    MR : ^MemRec absolute P;
    PSO : SegOfs absolute P;
  begin
    with MR^ do begin
      WriteLn(HexW(PSO.Ofst+1), ' ', HexPtr(meType), ' type def');
      DumpRawType(TP, meType, PSO.Segm);
    end;
  end;

  procedure DumpUnitName(TP : TpuPointer; P : Pointer);
    {-Dump a unit name symbol}
  var
    UP : ^UnitNameRec absolute P;
  begin
    with UP^ do begin
      Write(
        HexW(Ofs(usSegment)), ' ', HexW(usSegment), ' relative segment'^M^J,
        HexW(Ofs(usVersion)), ' ', HexW(usVersion), ' checksum'^M^J,
        HexW(Ofs(usNext)), ' ', HexW(usNext), ' ofs. of symbol for next unit');
      if usNext = 0 then
        WriteLn
      else
        WriteLn(' (', SymbolAt(TP, usNext), ')');
      Write(
        HexW(Ofs(usPrev)), ' ', HexW(usPrev), ' ofs. of symbol for previous unit');
      if usPrev = 0 then
        WriteLn
      else
        WriteLn(' (', SymbolAt(TP, usPrev), ')');
    end;
  end;

  procedure DumpSymbol(TP : TpuPointer; PHead : SymHeadPtr; DoCR : Boolean);
    {-Dump an individual symbol}
  var
    PSO : SegOfs absolute PHead;
    PTail : SymTailPtr;
    PTSO : SegOfs absolute PTail;
    Code : Char;
    Private : Boolean;
  begin
    with PHead^ do begin
      PTail := Pointer(PHead);
      Inc(PTSO.Ofst, 4+Length(SymName));
      Private := (Byte(SymCode) AND $80 <> 0);
      Code := Char(Byte(SymCode) AND $7F);
      if DoCR then
        WriteLn;
      Write(HexW(PSO.Ofst), ' ', HexW(LinkCode), ' link');
      if LinkCode = 0 then
        Writeln
      else
        WriteLn(' (-> ', SymbolAt(TP, LinkCode), ')');
      Write(HexW(PSO.Ofst+2), ' ''', Code, '''  ');
      if Private then Write('private ');
      WriteLn(SymbolClass(Code));
      WriteLn(HexW(PSO.Ofst+3), '':6, SymName);

      {dump rest of info}
      case Code of
        scConst    : DumpConstant(TP, PTail);
        scType     : DumpType(TP, PTail);
        scVar      : DumpVariable(TP, PTail);
        scProcFunc : DumpProcedure(TP, PTail);
        scUnit     : DumpUnitName(TP, PTail);
        scLabel    : DumpLabel(PTail);
        scSysProc,
        scSysFunc,
        scSysPcFc  : DumpSystemProc(PTail);
        scPort     : DumpPortInst(PTail);
        scMem      : DumpMemInst(TP, PTail);
        else         WriteLn('unknown code');
      end;
    end;
  end;

  procedure DumpSymbolTable(TP : TpuPointer; ScopePtr : Word);
    {-Dump a symbol table}
  var
    P : SymHeadPtr;
    PSO : SegOfs absolute P;
    UnitSeg, Entries, I, O : Word;
  begin
    P := Pointer(TP);
    UnitSeg := Seg(TP^);
    O := ScopePtr;
    Entries := MemW[UnitSeg:O] shr 1;
    if SymNest <> 0 then
      WriteLn('----------------------------------- Start local symbols level ',
        SymNest)
    else
      WriteLn('Symbol Table (not in order)'^M^J,
              'Ofst Link Name'^M^J'---- ---- ----');

    for I := 0 to Entries do begin
      Inc(O, 2);
      PSO.Ofst := O;
      PSO.Ofst := P^.LinkCode;
      while PSO.Ofst <> 0 do begin
        {dump this symbol}
        DumpSymbol(TP, P, True);
        {follow the links}
        PSO.Ofst := P^.LinkCode;
      end;
    end;
    WriteLn;
  end;

  procedure DumpFileTable(TP : TpuPointer);
    {-Dump a filename table}
  var
    P : ^NameRec;
    PSO : SegOfs absolute P;
    StopPoint : Word;
  begin
    P := Pointer(TP);
    PSO.Ofst := TP^.NamePtr;
    StopPoint := TP^.LineXlatePtr;
    WriteLn('File Table');
    WriteLn('Ofst Type     Indx Date/Time                 Name', ^M^J,
            '---- -------- ---- ------------------------- ------------');
    while (PSO.Ofst < StopPoint) do
      with P^ do begin
        WriteLn(HexW(PSO.Ofst), ' ', HexB(nClass), ' (',
          FileTypeTable[nClass], ') ', HexW(nIndex), ' ',
          HexL(nFileDate), ' (', FileDate2Str(nFileDate), ') ', nFileName);
        Inc(PSO.Ofst, Length(nFileName)+8);
      end;
    WriteLn;
  end;

  function GetFileName(TP : TpuPointer; I : Word) : string;
  var
    PF : ^NameRec;
    PFSO : SegOfs absolute PF;
  begin
    PFSO.Segm := SegOfs(TP).Segm;
    PFSO.Ofst := TP^.NamePtr+I;
    GetFileName := JustFilename(PF^.nFileName);
  end;

  procedure DumpLineXlateTable(TP : TpuPointer);
    {-Dump the line number translation table}
  type
    LineXlateRec =
      record
        FileIndex : Word;
        Delta, Lines : Word;
      end;
  var
    PLXR : ^LineXlateRec;
    PSO : SegOfs absolute PLXR;
    StopPoint, Count : Word;

    function RangeString(W1, W2 : Word) : string;
    var
      S : string;
    begin
      S := Long2Str(W1)+'-'+Long2Str(W2);
      RangeString := Pad(S, 9);
    end;

  begin
    PLXR := Pointer(TP);
    PSO.Ofst := TP^.LineXlatePtr;
    StopPoint := TP^.DebugPtr;
    WriteLn('Line Number Translation Table');
    WriteLn('Ofst File Delta Lines -> Filename.Ext Linear    Actual'^M^J,
            '---- ---- ----- -----    ------------ --------- ---------');
    Count := 0;
    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do begin
      with PLXR^ do
        WriteLn(HexW(PSO.Ofst), ' ', HexW(FileIndex), ' ', HexW(Delta), '  ',
                HexW(Lines), '  -> ', Pad(GetFileName(TP, FileIndex), 12), ' ',
                RangeString(Count+1, Count+Lines), ' ',
                RangeString(Delta+1, Lines+Delta));
      Inc(Count, PLXR^.Lines);
      Inc(PSO.Ofst, SizeOf(LineXlateRec));
    end;
    WriteLn;
  end;

  function FindProcNameAndEntryPoint(TP : TpuPointer; GroupOfst : Word;
                        NameOnly : Boolean; Var EntryPoint:Word) : Boolean;
    {-Find the name of a procedure given an offset in the group table}
  var
    PHead : SymHeadPtr;
    PHSO : SegOfs absolute PHead;
    PTail : SymTailPtr;
    PTSO : SegOfs absolute PTail;
    PR : ^ProcRec absolute PTail;
    PTP : ProcTablePtr;
    PO : ^ObjectRec;
    LHead : SymHeadPtr;
    LHSO : SegOfs absolute LHead;
    LTail : SymTailPtr;
    LTSO : SegOfs absolute LTail;
    FRP : ^FieldRec absolute LTail;
    PRP : ^ProcRec absolute LTail;
    Entries, I, J, O : Word;
  begin
    FindProcNameAndEntrypoint := False;
    EntryPoint := 0;
    PHSO.Segm := SegOfs(TP).Segm;
    PTSO.Segm := SegOfs(TP).Segm;
    SegOfs(PTP).Segm := SegOfs(TP).Segm;
    O := TP^.DScopePtr;
    Entries := MemW[SegOfs(TP).Segm:O] shr 1;
    for I := 0 to Entries do begin
      Inc(O, 2);
      PHSO.Ofst := O;
      PHSO.Ofst := PHead^.LinkCode;
      while PHSO.Ofst <> 0 do begin
        PTSO.Ofst := PHSO.Ofst+4+Length(PHead^.SymName);
        case Char(Byte(PHead^.SymCode) and $7F) of
          scProcFunc :
            with PR^ do
              if (pFlags and pInline = 0) then begin
                SegOfs(PTP).Ofst := TP^.ProcPtr+pIndex;
                if (PTP^.prGroup = GroupOfst) and (PTP^.prFlags and ptExtern = 0) then begin
                  if (pFlags and pExternal = 0) then
                    EntryPoint := PTP^.prOffset;
                  Write(PHead^.SymName);
                  if not NameOnly then
                    WriteLn(':'^M^J, CharStr('-', Length(PHead^.SymName)+1));
                  FindProcNameAndEntryPoint := True;
                  Exit;
                end;
              end;
          scType :
            if TypeIsObject(PHead) then begin
              PO := PTail^.tsType;
              SegOfs(PO).Segm := PTSO.Segm;
              SegOfs(PO).Ofst := SegOfs(PTail^.tsType).Ofst;

              LHSO.Segm := PTSO.Segm;
              LHSO.Ofst := PO^.tFieldPtr;
              while LHSO.Ofst <> 0 do begin
                LTSO.Segm := PTSO.Segm;
                LTSO.Ofst := LHSO.Ofst+4+Length(LHead^.SymName);

                {is it a field?}
                if Char(Byte(LHead^.SymCode) and $7F) = scVar then
                  LHSO.Ofst := FRP^.vNext
                {if not, it's a method}
                else begin
                  SegOfs(PTP).Ofst := TP^.ProcPtr+PRP^.pIndex;
                  if PTP^.prGroup = GroupOfst then begin
                    PR := Pointer(LTail);
                    if (PR^.pFlags and pExternal = 0) then
                      EntryPoint := PTP^.prOffset;
                    J := Length(PHead^.SymName)+Length(LHead^.SymName)+2;
                    Write(PHead^.SymName, '.', LHead^.SymName);
                    if not NameOnly then
                      WriteLn(':'^M^J, CharStr('-', J) );
                    FindProcNameAndEntryPoint := True;
                    Exit;
                  end;
                  LHSO.Ofst := PRP^.pType.tNext;
                end;
              end;
            end;
        end;

        {follow the links}
        PHSO.Ofst := PHead^.LinkCode;
      end;
    end;

    {No symbol exists, check if it's the initialization block}
    SegOfs(PTP).Ofst := TP^.ProcPtr; {Init block is always first entry, even if empty.}
    if PTP^.prGroup = GroupOfst then
      begin
        EntryPoint := PTP^.prOffset;
        Write('Initialization Block');
        FindProcNameAndEntryPoint := True;
      end
    else if not NameOnly then
      Write('??');
    if not NameOnly then
      WriteLn(':'^M^J, CharStr('-', 21) );
  end;

  procedure DumpLineTable(TP : TpuPointer);
    {-Dump a line number table}
  var
    P : Pointer;
    PSO : SegOfs absolute P;
    PG : ^GroupRec absolute P;
    PDR : ^DebugRec;
    PDRSO : SegOfs absolute PDR;
    PF : ^NameRec;
    PFSO : SegOfs absolute PF;
    I, L, O, StopPoint : Word;
    PB : ^Byte;
    PBSO : SegOfs absolute PB;
    PW : ^Word absolute PB;
    PS : ^String;
    PSSO : SegOfs absolute PS;
    EntryPoint : Word; {not used here}
  begin
    WriteLn('Debug Table (arranged by code group)');
    WriteLn('------------------------------------');
    {exit if there are no line numbers to do}
    if TP^.DebugPtr = TP^.UnitSize then begin
      WriteLn('(empty)');
      WriteLn;
      Exit;
    end;

    P := TP;
    PSSO.Segm := PSO.Segm;
    PSO.Ofst := TP^.GroupPtr;
    StopPoint := TP^.ConGrPtr;
    PDR := P;
    PF := P;
    WriteLn(
      'TPU  Proc File 1st  Code Line Stat'^M^J,
      'Ofst Indx Name Line Orgn Orgn Cnt  (Proc Name, Filename)'^M^J,
      '---- ---- ---- ---- ---- ---- ---- ----------------------------'^M^J,
      'Ofst Val = Abs Line'^M^J,
      '---- ---- ---- ----'^M^J);

    while (PSO.Ofst < StopPoint) do begin
      if (PG^.grDebugPtr <> $FFFF) then begin
        PDRSO.Ofst := TP^.DebugPtr+PG^.grDebugPtr;
        with PDR^ do begin
          PFSO.Ofst := TP^.NamePtr+deNamePtr;
          Write(
            HexW(PDRSO.Ofst), ' ',
            HexW(deProcPtr), ' ',
            HexW(deNamePtr), ' ',
            HexW(deFirstLine), ' ',
            HexW(deCodeOrg), ' ',
            HexW(deLineOrg), ' ',
            HexW(deLineCnt), ' (');
          if not FindProcNameAndEntryPoint(TP,
                 PSO.Ofst-TP^.GroupPtr, True, EntryPoint) then begin
            PSSO.Ofst := deProcPtr+3;
            if deProcPtr = 0 then
              Write('??')
            else
              Write(PS^);
          end;
          WriteLn(', ', JustFilename(PF^.nFileName), ')');

          L := deLineOrg;
          O := deCodeOrg;
          PB := @deStatTab;
          for I := 1 to deLineCnt do begin
            Write(HexW(Ofs(PB^)), ' ');
            if PB^ and $80 <> 0 then begin
              WriteLn(HexW(PW^), ' ', HexW(O), ' ', L);
              Inc(O, Swap(PW^) and $7FFF);
              Inc(PBSO.Ofst, 2);
            end
            else begin
              WriteLn(HexB(PB^), '   ', HexW(O), ' ', L);
              Inc(O, PB^);
              Inc(PBSO.Ofst);
            end;
            Inc(L);
          end;
        end;
      end
      else
        WriteLn('no line numbers for this group');

      WriteLn;
      Inc(PSO.Ofst, SizeOf(GroupRec));
    end;
  end;

  procedure DumpProcTable(TP : TpuPointer);
    {-Dump the procedure table}
  var
    P : ProcTablePtr;
    PSO : SegOfs absolute P;
    PGR : ^GroupRec;
    PGRSO : SegOfs absolute PGR;
    PS : ^String;
    PSSO : SegOfs absolute PS;
    PDR : ^DebugRec;
    PDRSO : SegOfs absolute PDR;
    StopPoint, EntryPoint : Word;

    function DynaLinkName(O : Word) : string;
    var
      LNP : ^LinkNameRec;
    begin
      LNP := Ptr(Seg(TP^), TP^.DynaLinkPtr+O);
      DynaLinkName := LNP^.kName;
    end;

  begin
    P := Pointer(TP);
    PSO.Ofst := TP^.ProcPtr;
    PDRSO.Segm := PSO.Segm;
    PGRSO.Segm := PSO.Segm;
    PSSO.Segm := PSO.Segm;
    StopPoint := TP^.GroupPtr;
    WriteLn('Procedure Table');
    WriteLn('Ofst ???? Flags Index Data Routine'^M^J'---- ---- ----- ----- ---- -----------');
    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do begin
      with P^ do begin
        Write(HexW(PSO.Ofst), ' ', HexW(prDunno), ' ', HexW(prFlags), '  ',
              HexW(prGroup), '  ', HexW(prOffset), ' ');
        if PSO.Ofst = TP^.ProcPtr then
          Write('Initialization Block')
        else if not FindProcNameByIndexFunc(TP, PSO.Ofst-TP^.ProcPtr) then
          if prOffset = 0 then begin
            if not FindProcNameAndEntryPoint(TP,
                   prGroup, True, EntryPoint) then begin
              PGRSO.Ofst := TP^.GroupPtr+prGroup;
              if PGR^.grDebugPtr = $FFFF then
                Write('??')
              else begin
                PDRSO.Ofst := TP^.DebugPtr+PGR^.grDebugPtr;
                PSSO.Ofst := PDR^.deProcPtr+3;
                if PDR^.deProcPtr = 0 then
                  Write('??')
                else
                  Write(PS^);
              end;
            end;
          end
          else
            Write('??');
          if prFlags and ptExport <> 0 then
            Write(' (export)')
          else if prFlags and ptExtern <> 0 then
            Write(' (external ''', DynaLinkName(prGroup), ''' index ',
                  prOffset, ')');
      end;
      WriteLn;
      Inc(PSO.Ofst, SizeOf(ProcTableRec));
    end;
    WriteLn;
  end;

  procedure DumpCodeGroupTable(TP : TpuPointer);
    {-Dump the code group table}
  var
    P : ^GroupRec;
    PSO : SegOfs absolute P;
    PS : ^String;
    PSSO : SegOfs absolute PS;
    PDR : ^DebugRec;
    PDRSO : SegOfs absolute PDR;
    StopPoint, EntryPoint : Word;
  begin
    P := Pointer(TP);
    PSO.Ofst := TP^.GroupPtr;
    PSSO.Segm := PSO.Segm;
    PDRSO.Segm := PSO.Segm;
    StopPoint := TP^.ConGrPtr;
    WriteLn('Code Group Table');
    WriteLn('Ofst Code Size Fixup Debug Routine'^M^J'---- ---- ---- ----- ----- -----------');

    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do
      with P^ do begin
        Write(HexW(PSO.Ofst), ' ',
          HexW(grOrigin), ' ',
          HexW(grCodeSize), ' ',
          HexW(grFixupSize), '  ',
          HexW(grDebugPtr), '  ');
          if not FindProcNameAndEntryPoint(TP,
                 PSO.Ofst-TP^.GroupPtr, True, EntryPoint) then
            if grDebugPtr = $FFFF then
              Write('??')
            else begin
              PDRSO.Ofst := TP^.DebugPtr+grDebugPtr;
              PSSO.Ofst := PDR^.deProcPtr+3;
              if PDR^.deProcPtr = 0 then
                Write('??')
              else
                Write(PS^);
            end;
        WriteLn;
        Inc(PSO.Ofst, SizeOf(GroupRec));
      end;
    WriteLn;
  end;

  procedure DumpConstGroupTable(TP : TpuPointer);
    {-Dump a const group table}
  var
    P : ^DataGroupRec;
    PSO : SegOfs absolute P;
    ORP : ^ObjectRec;
    ORPSO : SegOfs absolute ORP;
    StopPoint : Word;
  begin
    P := Pointer(TP);
    ORPSO.Segm := PSO.Segm;
    PSO.Ofst := TP^.ConGrPtr;
    StopPoint := TP^.DatGrPtr;
    WriteLn('Const Group Table');
    WriteLn('Ofst Orgn Size Fixup Debug What it is'^M^J'---- ---- ---- ----- ----- ------------');

    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do
      with P^ do begin
        Write(HexW(PSO.Ofst), ' ',
          HexW(cdOrigin), ' ',
          HexW(cdSize), ' ',
          HexW(cdFixupSize), '  ',
          HexW(cdDebugPtr), '  ');
        if cdDebugPtr <> 0 then begin
          ORPSO.Ofst := cdDebugPtr;
          if (ORP^.tClass = tObject) and (ORP^.tNameOfs <> 0) then begin
            Write(SymbolAt(TP, ORP^.tNameOfs), '''s ');
            if ORP^.tMethGroup = PSO.Ofst-TP^.ConGrPtr then
              Write('VMT')
            else if ORP^.tDMTGroup = PSO.Ofst-TP^.ConGrPtr then
              Write('DMT')
            else
              Write('??');
          end
          else
            Write('??');
        end
        else
          Write('??');
        WriteLn;
        Inc(PSO.Ofst, SizeOf(DataGroupRec));
      end;
    WriteLn;
  end;

  procedure DumpDataGroupTable(TP : TpuPointer);
    {-Dump a data group table}
  var
    P : ^DataGroupRec;
    PSO : SegOfs absolute P;
    StopPoint : Word;
  begin
    P := Pointer(TP);
    PSO.Ofst := TP^.DatGrPtr;
    StopPoint := TP^.DynaLinkPtr;
    WriteLn('Data Group Table');
    WriteLn('Ofst Orgn Size Fixup Debug'^M^J'---- ---- ---- ----- -----');

    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do
      with P^ do begin
        WriteLn(HexW(PSO.Ofst), ' ',
          HexW(cdOrigin), ' ',
          HexW(cdSize), ' ',
          HexW(cdFixupSize), '  ',
          HexW(cdDebugPtr));
        Inc(PSO.Ofst, SizeOf(DataGroupRec));
      end;
    WriteLn;
  end;

  procedure DumpDynaLinkNamesTable(TP : TpuPointer);
    {-Dump a link names table}
  var
    P : ^LinkNameRec;
    PSO : SegOfs absolute P;
    StopPoint : Word;
  begin
    P := Pointer(TP);
    PSO.Ofst := TP^.DynaLinkPtr;
    StopPoint := TP^.LinkPtr;
    WriteLn('DLL Link Names Table');
    WriteLn('Ofst Segm Xrefs Name'^M^J'---- ---- ----- --------------------');

    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do
      with P^ do begin
        WriteLn(HexW(PSO.Ofst), ' ', HexW(kSegment), ' ', HexW(kXrefCnt), '  ', kName);
        Inc(PSO.Ofst, 5+Length(kName));
      end;
    WriteLn;
  end;

  procedure DumpLinkNamesTable(TP : TpuPointer);
    {-Dump a link names table}
  var
    P : ^LinkNameRec;
    PSO : SegOfs absolute P;
    StopPoint : Word;
  begin
    P := Pointer(TP);
    PSO.Ofst := TP^.LinkPtr;
    StopPoint := TP^.NamePtr;
    WriteLn('Link Names Table');
    WriteLn('Ofst Segm Xrefs Name'^M^J'---- ---- ----- --------------------');

    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do
      with P^ do begin
        WriteLn(HexW(PSO.Ofst), ' ', HexW(kSegment), ' ', HexW(kXrefCnt), '  ', kName);
        Inc(PSO.Ofst, 5+Length(kName));
      end;
    WriteLn;
  end;

  function TypeIsObject(PHead : SymHeadPtr) : Boolean;
    {-Return True if P points to an object}
  var
    PHSO : SegOfs absolute PHead;
    PTail : SymTailPtr absolute PHead;
    TRP : ^TypeRec absolute PTail;
    RTP : RawTypePtr;
  begin
    TypeIsObject := False;
    Inc(PHSO.Ofst, 4+Length(PHead^.SymName));
    if SegOfs(TRP^.tsType).Segm = CurrentUnitSymbol then begin
      with SegOfs(RTP) do begin
        Segm := PHSO.Segm;
        Ofst := SegOfs(TRP^.tsType).Ofst;
      end;
      TypeIsObject := RTP^.tClass = tObject;
    end;
  end;

  procedure DumpHexAt(P : Pointer; Bytes : Word); {6.00 - entire proc}
    {-Dump info as hex}
  type
    Data = array[0..65520] of Byte;
    DataPtr = ^Data;
  var
    I : Word;
    OutSt : string[80];
    Save : array[0..$F] of Byte;

    procedure SendOut(I : word; B : byte);
    var
      N : byte;
      S : string[16];
      C : char;
    begin
      N := I and $000F;
      if N = 0 then
        OutSt := HexW(I) + ' ';
      OutSt := OutSt + ' ' + HexB(B);
      Save[N] := b;
      if N = $F then begin
        S[0] := #16;
        for N := 0 to $F do begin
          C := Char(Save[N]);
          if C < ' ' then
            S[N+1] := '.'
          else
            S[N+1] := C;
        end;
        WriteLn(OutSt, '  |  ', S);
      end;
    end;

    procedure CloseOut(I : Word);
    var
      J, N : Byte;
      S : string[16];
      C : Char;
    begin
      J := I and $000F;
      if J < $F then begin
        for N := J+1 to $F do
          OutSt := OutSt + '   ';
        S[0] := Char(J+1);
        for N := 0 to J do begin
          C := Char(Save[N]);
          if C < ' ' then
            S[N+1] := '.'
          else
            S[N+1] := C;
        end;
        WriteLn(OutSt, '  |  ', S);
      end;
    end;

  begin
    WriteLn('--------------------------------------------------------------------------');
    WriteLn('      00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F     0123456789ABCDEF');
    WriteLn('--------------------------------------------------------------------------');
    for I := 0 to Pred(Bytes) do
      SendOut(I, DataPtr(P)^[I]);
    CloseOut(I);
  end;

  procedure DumpConstData(TP : TpuPointer);
    {-Dump constant data}
  begin
    WriteLn('Constant Data');
    with TP^ do
      if ConstSize = 0 then begin
        WriteLn('-------------');
        WriteLn('(none)');
      end
      else
        DumpHexAt(Ptr(CodeSeg+Paras(CodeSize), 0),ConstSize);
    WriteLn;
  end;

  procedure DumpBrowserData(TP : TpuPointer);
    {-Dump browser data}
  type
    bSymRec =
      record
        Sym, Line : Word;
      end;
    bStructRec =
      record
        Start, Stop, TypeDef, LinkTabOfs : Word;
      end;
    LineXlateRec =
      record
        FileIndex : Word;
        Delta, Lines : Word;
      end;
  var
    PSR : ^bSymRec;
    PStR : ^bStructRec absolute PSR;
    PSO : SegOfs absolute PSR;
    LNP : ^LinkNameRec;
    LNPSO : SegOfs absolute LNP;
    LC, StopPoint : Word;
    UN : ^string;

    function LinkName(O : Word) : string;
    var
      LNP : ^LinkNameRec;
    begin
      LNP := Ptr(Seg(TP^), TP^.LinkPtr+O);
      LinkName := LNP^.kName;
    end;

    function LinearToActual(Line1, Line2 : Word) : string;
    var
      S : string;
      PLXR : ^LineXlateRec;
      PSO : SegOfs absolute PLXR;
      Count : Word;
    begin
      PLXR := Pointer(TP);
      PSO.Ofst := TP^.LineXlatePtr;
      Count := 0;
      repeat
        if Line1 <= (Count+PLXR^.Lines) then
          with PLXR^ do begin
            S := StLocase(GetFilename(TP, PLXR^.FileIndex))+' ';
            if Line2 = Line1 then
              LinearToActual := S+Long2Str(Delta+Line1-Count)
            else
              LinearToActual :=
                S+Long2Str(Delta+Line1-Count)+'-'+Long2Str(Delta+Line2-Count);
            Exit;
          end;
        Inc(Count, PLXR^.Lines);
        Inc(PSO.Ofst, SizeOf(LineXlateRec));
      until False;
    end;

  begin
    if TP^.BrowseSize = 0 then begin
      WriteLn('Browser Data'^M^J'------------'^M^J'(none)'^M^J);
      Exit;
    end;

    PSR := Ptr(BaseSeg+Paras(TP^.UnitSize), 0);
    WriteLn('Browser Data: Symbols'^M^J'---------------------');
    WriteLn('Ofst Sym  Line SymbolName (location)'^M^J'---- ---- ---- ---------------------');
    StopPoint := TP^.BrowserXrefs;
    while PSO.Ofst < StopPoint do begin
      with PSR^ do
        WriteLn(
          HexW(PSO.Ofst), ' ', HexW(Sym), ' ', HexW(Line), ' ', SymbolAt(TP, Sym),
          ' (', LinearToActual(Line, Line), ')');
      Inc(PSO.Ofst, SizeOf(bSymRec));
    end;
    WriteLn;

    WriteLn('Browser Data: Cross-References'^M^J'------------------------------');
    WriteLn('Ofst Sym  Line Unit:SymbolName (location)'^M^J'---- ---- ---- --------------------------');
    StopPoint := TP^.BrowserStructs;
    LNP := Ptr(Seg(TP^), TP^.LinkPtr);
    while LNP^.kXrefCnt = 0 do
      Inc(LNPSO.Ofst, 5+Length(LNP^.kName));
    LC := LNP^.kXrefCnt div SizeOf(bSymRec);
    UN := Ptr(Seg(TP^), TP^.UsesPtr+3);
    while PSO.Ofst < StopPoint do begin
      with PSR^ do begin
        Write(
          HexW(PSO.Ofst), ' ', HexW(Sym), ' ', HexW(Line), ' ', LNP^.kName, ':');
        if LNP^.kName <> UN^ then
          Write('??')
        else
          Write(SymbolAt(TP, Sym));
        WriteLn(' (', LinearToActual(Line, Line), ')');
      end;
      Inc(PSO.Ofst, SizeOf(bSymRec));
      if PSO.Ofst < StopPoint then begin
        Dec(LC);
        if LC = 0 then begin
          repeat
            Inc(LNPSO.Ofst, 5+Length(LNP^.kName));
          until LNP^.kXrefCnt <> 0;
          LC := LNP^.kXrefCnt div SizeOf(bSymRec);
        end;
      end;
    end;
    WriteLn;

    WriteLn('Browser Data: Structures'^M^J'------------------------');
    WriteLn('Ofst Start Stop Type LT ofs Unit:TypeName (location)'^M^J,
            '---- ----- ---- ---- ------ ------------------------');
    StopPoint := TP^.BrowseSize;
    while PSO.Ofst < StopPoint do begin
      with PStR^ do
        WriteLn(
          HexW(PSO.Ofst), ' ', HexW(Start), '  ', HexW(Stop),
               ' ', HexW(TypeDef), '  ', HexW(LinkTabOfs), '  ',
               LinkName(LinkTabOfs), ':',
               FindTypeNameFunc(TP, Ptr(CurrentUnitSymbol, TypeDef), True),
               ' (', LinearToActual(Start, Stop), ')');
      Inc(PSO.Ofst, SizeOf(bStructRec));
    end;
    WriteLn;
  end;

  procedure DumpCodeBlocks(TP : TpuPointer);
    {-Dump the code blocks}
  var
    P : Pointer;
    PSO : SegOfs absolute P;
    GP : ^GroupRec absolute P;
    IP, StopPoint : Word;
    B : Boolean;
    EntryPoint : Word;
  begin
    WriteLn('Code Blocks'^M^J'-----------');

    P := TP;
    PSO.Ofst := TP^.GroupPtr;
    StopPoint := TP^.ConGrPtr;
    CodeSeg := BaseSeg+Paras(TP^.UnitSize)+Paras(TP^.BrowseSize);
    IP := 0;
    if (PSO.Ofst = StopPoint) then
      WriteLn('(none)')
    else while (PSO.Ofst < StopPoint) do begin
      WriteLn;
      B := FindProcNameAndEntryPoint(TP, PSO.Ofst-TP^.GroupPtr,
        False, EntryPoint);

      if EntryPoint > 0 then begin
        Writeln('CS Constant Data');
        DumpHexAt(Ptr(CodeSeg, IP), EntryPoint);
        Writeln;
      end;
      Unassemble(IP+EntryPoint, GP^.grCodeSize-EntryPoint, False,
                Ptr(CodeSeg, IP+EntryPoint), Output);
      Inc(IP, GP^.grCodeSize);
      Inc(PSO.Ofst, SizeOf(GroupRec));
    end;
    WriteLn;
  end;

  procedure DumpCodeFixupTable(TP : TpuPointer);
    {-Dump the code fixup table}
  var
    FT : ^FixupRec;
    P : Pointer;
    PSO : SegOfs absolute P;
    PG : ^GroupRec absolute P;
    StopPoint,
    I, Entries : Word;
  begin
    WriteLn('Code Fixup Table'^M^J'----------------');
    WriteLn('TPU  Fix  Fix  Fix  Fix');
    WriteLn('Ofst Type Indx Ofst Locn'^M^J'---- ---- ---- ---- ----');

    with TP^ do
      RelSeg := CodeSeg+Paras(CodeSize)+Paras(ConstSize);
    FT := Ptr(RelSeg, 0);
    P := TP;
    PSO.Ofst := TP^.GroupPtr;
    StopPoint := TP^.ConGrPtr;

    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)'^M^J)
    else while (PSO.Ofst < StopPoint) do begin
      Entries := PG^.grFixupSize div SizeOf(FixupRec);
      if Entries = 0 then
        WriteLn('(no entries in group)')
      else for I := 1 to Entries do begin
        with FT^ do
          WriteLn(
            HexW(SegOfs(FT).Ofst), ' ', HexW(fxFlags), ' ', HexW(fxIndex), ' ',
            HexW(fxOffset), ' ', HexW(fxLocation));
        Inc(SegOfs(FT).Ofst, SizeOf(FixupRec));
      end;
      Inc(PSO.Ofst, SizeOf(GroupRec));
      WriteLn;
    end;
  end;

  procedure DumpConstFixupTable(TP : TpuPointer);
    {-Dump the constant date fixup table}
  var
    FT : ^FixupRec;
    P : Pointer;
    PSO : SegOfs absolute P;
    PG : ^GroupRec absolute P;
    StopPoint,
    I, Entries : Word;
  begin
    WriteLn('Const Fixup Table'^M^J'-----------------');
    WriteLn('TPU  Fix  Fix  Fix  Fix');
    WriteLn('Ofst Type Indx Ofst Locn'^M^J'---- ---- ---- ---- ----');

    with TP^ do
      RelSeg := CodeSeg+Paras(CodeSize)+Paras(ConstSize)+Paras(FixupSize);
    FT := Ptr(RelSeg, 0);
    P := TP;
    PSO.Ofst := TP^.ConGrPtr;
    StopPoint := TP^.DatGrPtr;

    if (PSO.Ofst = StopPoint) then
      WriteLn('(empty)')
    else while (PSO.Ofst < StopPoint) do begin
      Entries := PG^.grFixupSize div SizeOf(FixupRec);
      if Entries = 0 then
        WriteLn('(no entries in group)')
      else for I := 1 to Entries do begin
        with FT^ do
          WriteLn(
            HexW(SegOfs(FT).Ofst), ' ', HexW(fxFlags), ' ', HexW(fxIndex), ' ',
            HexW(fxOffset), ' ', HexW(fxLocation));
        Inc(SegOfs(FT).Ofst, SizeOf(FixupRec));
      end;
      Inc(PSO.Ofst, SizeOf(GroupRec));
       WriteLn;
    end;
  end;

  function UnitType(Header : TpuPointer) : string;
    {-Returns 'DOS', 'DPMI', or 'Windows'}
  begin
    case Header^.UnitFlags and ufWindows of
      ufDOS : UnitType := 'DOS';
      ufDPMI : UnitType := 'DPMI';
      ufWindows : UnitType := 'Windows';
    end;
  end;

  procedure DumpTpuHeader(Header : TpuPointer);
    {-Dump a TPU header}
  var
    UN : ^string;
    i  : Word;

  begin
    with Header^ do begin
      UN := Ptr(Seg(Header^), Header^.UsesPtr+3);
      WriteLn(^M^J'.TPU File Header - ', SizeOf(Header^), ' bytes');
      WriteLn('Ofst Value  Meaning'^M^J'---- ------ ---------------------');
      WriteLn('0000 "', TPUsig , '" 4-byte TPU signature');
      DumpWord(ofs(NextUnit)   , NextUnit,    'segment in memory for next unit');
      DumpWord(ofs(NextLibrary), NextLibrary, 'segment in memory for next library');
      DumpWord(ofs(UsesPtr)    , UsesPtr,     'offset to name of current unit ('+UN^+')');
      DumpWord(ofs(ScopePtr)   , ScopePtr,    'offset to scope table');
      DumpWord(ofs(ProcPtr)    , ProcPtr,     'offset to procedure table');
      DumpWord(ofs(GroupPtr)   , GroupPtr,    'offset to code group table');
      DumpWord(ofs(ConGrPtr)   , ConGrPtr,    'offset to const group table');
      DumpWord(ofs(DatGrPtr)   , DatGrPtr,    'offset to data group table');
      DumpWord(ofs(DynaLinkPtr), DynaLinkPtr, 'offset to DLL link names table');
      DumpWord(ofs(LinkPtr)    , LinkPtr,     'offset to link names table');
      DumpWord(ofs(NamePtr)    , NamePtr,     'offset to filename table');
      DumpWord(ofs(LineXlatePtr),LineXlatePtr,'offset to line number translation table');
      DumpWord(ofs(DebugPtr)   , DebugPtr,    'offset to debug table');
      DumpWord(ofs(UnitSize)   , UnitSize,    'unit (symbol table) size');
      DumpWord(ofs(BrowseSize) , BrowseSize,  'size of browser data');
      DumpWord(ofs(CodeSize)   , CodeSize,    'total code (bytes)');
      DumpWord(ofs(ConstSize)  , ConstSize,   'initialized data (bytes)');
      DumpWord(ofs(FixupSize)  , FixupSize,   'size of fixup table');
      DumpWord(ofs(ConFixSize) , ConFixSize,  'size of the constant fixup table');
      DumpWord(ofs(DataSize)   , DataSize,    'uninitialized data (bytes)');
      DumpWord(ofs(DScopePtr)  , DScopePtr,   'debug scope pointer');
      DumpWord(ofs(UnitFlags)  , UnitFlags,
                               '$N'+PlusMinus[UnitFlags and ufNPlus <> 0]+
                               ', $O'+PlusMinus[UnitFlags and ufOPlus <> 0]+
                               ', $G'+PlusMinus[UnitFlags and ufGPlus <> 0]+
                               ', '+UnitType(Header));
      DumpWord(ofs(LastObjectPtr), LastObjectPtr, 'last object in linked list ('+
        GetObjectLink(Header, LastObjectPtr)+')');
      DumpWord(ofs(BrowserXrefs), BrowserXrefs,   'offset in browser data for cross references');
      DumpWord(ofs(BrowserStructs), BrowserStructs, 'offset in browser data for structures');
      for i := 1 to SizeOf(Private) div 2 do
        DumpWord(ofs(Private[i]), Private[i],  '??');
    end;
    WriteLn;
  end;

  function FindSystemSeg(TPU : TpuPointer) : Word;
    {-Find the 'segment' for the SYSTEM unit}
  var
    SHP : SymHeadPtr;
    SHPSO : SegOfs absolute SHP;
    UNP : ^UnitNameRec;
    UNPSO : SegOfs absolute UNP;
  begin
    SHP := Ptr(SegOfs(TPU).Segm, TPU^.UsesPtr);
    UNPSO.Segm := SHPSO.Segm;
    repeat
      UNPSO.Ofst := SHPSO.Ofst;
      Inc(UNPSO.Ofst, Length(SHP^.SymName)+4);
      if StUpcase(SHP^.SymName) = 'SYSTEM' then begin
        FindSystemSeg := UNPSO.Ofst;
        Exit
      end;

      SHPSO.Ofst := UNP^.usNext;
    until SHPSO.Ofst = 0;
    FindSystemSeg := $FFFF;
  end;

  procedure DumpRawData;
    {-Dump a 6.0 TPU or TPL file}
  const
    FormFeed = ^L;
  var
    P : TpuPointer;
    PSO : SegOfs absolute P;
    SHP : SymHeadPtr;
    OK : Boolean;
  begin
    {base segment is segment for first TPU}
    P := Buffer;
    repeat
      BaseSeg := PSO.Segm;

      {read the symbol table}
      ReadSymbolTableProc(P, P^.DScopePtr);

      {dump the header}
      DumpTpuHeader(P);

      {get offset of current unit symbol}
      CurrentUnitSymbol := P^.UsesPtr;
      SHP := Ptr(PSO.Segm, CurrentUnitSymbol);
      Inc(CurrentUnitSymbol, Length(SHP^.SymName)+4);
      SystemSeg := FindSystemSeg(P);

      if P^.DScopePtr <> P^.ScopePtr then begin
        {dump the interface's hash table}
        OK := DumpHashTable(P, P^.ScopePtr);

        {dump the implementation's hash table}
        OK := DumpHashTable(P, P^.DScopePtr);

        {dump the symbol table}
        DumpSymbolTable(P, P^.DScopePtr);
      end
      else begin
        {dump the interface's hash table}
        OK := DumpHashTable(P, P^.ScopePtr);

        {dump the symbol table}
        DumpSymbolTable(P, P^.ScopePtr);
      end;

      {dump the procedure table}
      DumpProcTable(P);

      {dump the code group table}
      DumpCodeGroupTable(P);

      {dump the const group table}
      DumpConstGroupTable(P);

      {dump the data group table}
      DumpDataGroupTable(P);

      {dump the DLL link names table}
      DumpDynaLinkNamesTable(P);

      {dump the link names table}
      DumpLinkNamesTable(P);

      {dump the file table}
      DumpFileTable(P);

      {dump the line number translation table}
      DumpLineXlateTable(P);

      {dump the line table}
      DumpLineTable(P);

      {dump browser data}
      DumpBrowserData(P);

      {disassemble the code}
      DumpCodeBlocks(P);

      {dump constant data}
      DumpConstData(P);

      {dump the code fixup table}
      DumpCodeFixupTable(P);

      {dump the const fixup table}
      DumpConstFixupTable(P);

      {write a form feed}
      WriteLn(FormFeed);

      {point to next TPU, if any}
      with P^ do
        Inc(PSO.Segm,
            Paras(UnitSize)+Paras(CodeSize)+Paras(FixupSize)+Paras(ConstSize));
    until (P^.TPUsig <> SigForTPU70);
  end;

end.
